package com.emarcotte.testrio;

import java.util.ArrayList;

import javax.media.opengl.GL;

import com.emarcotte.glfoo.math.Vector3;

/**
 * Mesh coating over the joints.
 * 
 * @author Eugene Marcotte
 */

public class MD5Mesh {
	private ArrayList<MD5Weight> weights;
	private ArrayList<MD5Vertex> verts;
	private double[] vertsData;
	private double[] normalData;
	private Vector3[] bindPositionNormals;
	private double[] uvData;
	private int[] faceData;

	/**
	 * Setup the weights array, with a given number of elements.
	 * 
	 * @param numWeights
	 *            The number of weights to make storage for.
	 */
	public void setNumWeights(int numWeights) {
		weights = new ArrayList<MD5Weight>(numWeights);
	}

	/**
	 * Setup the vertices array, with a given number of elements.
	 * 
	 * @param numVertices
	 *            The number of verticies
	 */
	public void setNumVertices(int numVertices) {
		verts = new ArrayList<MD5Vertex>(numVertices);
		vertsData = new double[numVertices * 3];
		normalData = new double[numVertices * 3];
		bindPositionNormals = new Vector3[numVertices];
		uvData = new double[numVertices * 2];
	}

	/**
	 * Setup the triangles array, with a given number of elements.
	 * 
	 * @param numTriangles
	 *            The number of triangles to make storage for.
	 */
	public void setNumTriangles(int numTriangles) {
		faceData = new int[numTriangles * 3];
	}

	public void addVertex(int vertId, double textureU, double textureV,
			int weightIndex, int weightElem) {
		MD5Vertex v = new MD5Vertex(vertId, weightIndex, weightElem);
		verts.add(vertId, v);
		uvData[vertId * 2] = textureU;
		uvData[vertId * 2 + 1] = textureV;
	}

	/**
	 * Add a triangle to the mesh.
	 * 
	 * @param index
	 *            The ID of the triangle.
	 * @param t
	 *            The Triangle to add.
	 */
	public void addTriangle(int index, int vert1, int vert2, int vert3) {
		faceData[index * 3] = vert1;
		faceData[index * 3 + 1] = vert2;
		faceData[index * 3 + 2] = vert3;
	}

	public void addWeight(int index, double x, double y, double z,
			double weightValue, int jointIndex) {
		MD5Weight w = new MD5Weight(x, y, z, weightValue, jointIndex);
		weights.add(index, w);
	}

	public ArrayList<MD5Vertex> getVerts() {
		return verts;
	}

	public ArrayList<MD5Weight> getWeights() {
		return weights;
	}

	public void draw(GL gl) {
		gl.glBegin(GL.GL_TRIANGLES);
		for (int i = 0; i < faceData.length; i += 3) {
			int v2 = faceData[i + 1];
			int v1 = faceData[i];
			int v3 = faceData[i + 2];

			gl.glTexCoord2d(getU(v2), getV(v2));
			gl.glNormal3d(getNX(v2), getNY(v2), getNZ(v2));
			gl.glVertex3d(getX(v2), getY(v2), getZ(v2));

			gl.glTexCoord2d(getU(v1), getV(v1));
			gl.glNormal3d(getNX(v1), getNY(v1), getNZ(v1));
			gl.glVertex3d(getX(v1), getY(v1), getZ(v1));

			gl.glTexCoord2d(getU(v3), getV(v3));
			gl.glNormal3d(getNX(v3), getNY(v3), getNZ(v3));
			gl.glVertex3d(getX(v3), getY(v3), getZ(v3));
		}
		gl.glEnd();

	}

	public void drawNormals(GL gl) {
		gl.glBegin(GL.GL_LINES);
		for (int i = 0; i < verts.size(); i++) {
			gl.glVertex3d(getX(i), getY(i), getZ(i));
			gl.glVertex3d(getX(i) + getNX(i), getY(i) + getNY(i), getZ(i)
					+ getNZ(i));
		}
		gl.glEnd();
	}

	public void computeBindPositionNormals() {
		// the
		double[] normalTemp = new double[3 * verts.size()];
		int[] fCount = new int[verts.size()];
		Vector3 n1 = new Vector3();
		Vector3 n2 = new Vector3();
		Vector3 normal = new Vector3();

		// determine each faces' contribution to vertices normals.
		for (int i = 0; i < faceData.length; i += 3) {
			Vector3 v1 = getPos(faceData[i]);
			Vector3 v2 = getPos(faceData[i + 1]);
			Vector3 v3 = getPos(faceData[i + 2]);

			// figure out face normal
			v2.subtract(v1, n1);
			v3.subtract(v1, n2);
			n1.normalize();
			n2.normalize();
			n2.crossProduct(n1, normal);
			normal.normalize();

			// add to each vertices' array for storage and keep track of the
			// count.
			addNormalData(normalTemp, faceData[i], fCount, normal);
			addNormalData(normalTemp, faceData[i + 1], fCount, normal);
			addNormalData(normalTemp, faceData[i + 2], fCount, normal);
		}

		// Average all the normal values for each vertex.
		for (int i = 0; i < normalTemp.length / 3; i++) {
			averageNormals(normalTemp, fCount[i], i);
		}
	}

	/**
	 * Given an array of all sums of components of normal data for all vertices,
	 * take an average and normalize it. The result is a vertex normal. This
	 * method stores the vertex normal in a bind position array and the normal
	 * data array.
	 * 
	 * @param normalTemp
	 *            The sum of components array.
	 * @param count
	 *            The number of faces which contribute to the sum of components.
	 * @param i
	 *            The vertex that is being worked on.
	 */
	private void averageNormals(double[] normalTemp, int count, int i) {
		// create vector with averaged normals
		Vector3 n = new Vector3(normalTemp[i * 3] / count,
				normalTemp[i * 3 + 1] / count, normalTemp[i * 3 + 2] / count);

		// normalize result
		n.normalize();

		// keep track of bind position normal
		bindPositionNormals[i] = n;

		// set initial normal data
		normalData[i * 3] = n.i;
		normalData[i * 3 + 1] = n.j;
		normalData[i * 3 + 2] = n.k;

	}

	private static void addNormalData(double[] normalTemp, int id,
			int[] fCount, Vector3 n) {

		normalTemp[id * 3] += n.getI();
		normalTemp[id * 3 + 1] += n.getJ();
		normalTemp[id * 3 + 2] += n.getK();
		fCount[id]++;
	}

	public double getU(int id) {
		return uvData[id * 2];
	}

	public double getV(int id) {
		return uvData[id * 2 + 1];
	}

	public double getX(int id) {
		return vertsData[id * 3];
	}

	public double getY(int id) {
		return vertsData[id * 3 + 1];
	}

	public double getZ(int id) {
		return vertsData[id * 3 + 2];
	}

	public void setVertData(int id, double x, double y, double z) {
		vertsData[id * 3] = x;
		vertsData[id * 3 + 1] = y;
		vertsData[id * 3 + 2] = z;
	}

	/**
	 * Use with caution. Creates a new vector. modifying this vector will not
	 * modify original data.
	 * 
	 * @param id
	 * @return
	 */
	public Vector3 getPos(int id) {
		return new Vector3(vertsData[id * 3], vertsData[id * 3 + 1],
				vertsData[id * 3 + 2]);
	}

	public double getNX(int id) {
		return normalData[id * 3];
	}

	public double getNY(int id) {
		return normalData[id * 3 + 1];
	}

	public double getNZ(int id) {
		return normalData[id * 3 + 2];
	}

	public void update(MD5Frame frame) {
		for (MD5Vertex v : verts) {
			v.recalculatePosition(this, weights, frame);
		}
	}

	public double[] getTexCoords() {
		return uvData;
	}

	public double[] getNormals() {
		return normalData;
	}

	public double[] getPositions() {
		return vertsData;
	}

	public int[] getIndices() {
		return faceData;
	}

	public Vector3 getNormal(int index) {
		return bindPositionNormals[index];
	}

	public void setNormalData(int index, Vector3 rotation) {
		normalData[index * 3] = rotation.i;
		normalData[index * 3 + 1] = rotation.j;
		normalData[index * 3 + 2] = rotation.k;

	}

}
