package utils;

import java.awt.color.ColorSpace;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.HashMap;
import java.util.Hashtable;

import javax.imageio.ImageIO;
import javax.media.opengl.GL;

/**
 * A utility class to load textures for JOGL
 * 
 * @author Kevin Glass
 */
public class TextureManager {
	private static TextureManager instance = new TextureManager();

	private HashMap<String, Texture> table;

	private static final ColorModel glAlphaColorModel = new ComponentColorModel(
			ColorSpace.getInstance(ColorSpace.CS_sRGB),
			new int[] { 8, 8, 8, 8 }, true, false,
			ComponentColorModel.TRANSLUCENT, DataBuffer.TYPE_BYTE);

	private static final ColorModel glColorModel = new ComponentColorModel(
			ColorSpace.getInstance(ColorSpace.CS_sRGB),
			new int[] { 8, 8, 8, 0 }, false, false, ComponentColorModel.OPAQUE,
			DataBuffer.TYPE_BYTE);

	private TextureManager() {
		table = new HashMap<String, Texture>();

	}

	private int createTextureID(GL gl) {
		int[] tmp = new int[1];
		gl.glGenTextures(1, tmp, 0);
		return tmp[0];
	}

	public static TextureManager getInstance() {
		return instance;
	}

	public Texture getTexture(String resourceName) {
		return table.get(resourceName);
	}

	public Texture getTexture(GL gl, String resourceName) throws IOException {

		Texture tex = getTexture(resourceName);

		if (tex != null) {
			return tex;
		}

		tex = loadTexture(gl, resourceName, GL.GL_TEXTURE_2D, GL.GL_RGBA,
				GL.GL_LINEAR, GL.GL_LINEAR, true);

		table.put(resourceName, tex);

		return tex;
	}

	private void uploadTexture(GL gl, Texture t, int target,
			int dstPixelFormat, int minFilter, int magFilter, boolean wrap) {

		int srcPixelFormat = GL.GL_RGB;

		if (t.getImage().getColorModel().hasAlpha()) {
			srcPixelFormat = GL.GL_RGBA;
		}

		byte[] data = ((DataBufferByte) t.getImage().getRaster()
				.getDataBuffer()).getData();

		ByteBuffer imageBuffer = ByteBuffer.allocateDirect(data.length);
		imageBuffer.order(ByteOrder.nativeOrder());
		imageBuffer.put(data, 0, data.length);
		imageBuffer.rewind();

		int wrapMode;
		if (wrap)
			wrapMode = GL.GL_REPEAT;
		else
			wrapMode = GL.GL_CLAMP;

		if (target == GL.GL_TEXTURE_2D) {
			gl.glTexParameteri(target, GL.GL_TEXTURE_WRAP_S, wrapMode);
			gl.glTexParameteri(target, GL.GL_TEXTURE_WRAP_T, wrapMode);
			gl.glTexParameteri(target, GL.GL_TEXTURE_MIN_FILTER, minFilter);
			gl.glTexParameteri(target, GL.GL_TEXTURE_MAG_FILTER, magFilter);
		}

		gl.glTexImage2D(target, 0, dstPixelFormat, t.getWidth(), t.getWidth(),
				0, srcPixelFormat, GL.GL_UNSIGNED_BYTE, imageBuffer);
	}

	private Texture loadTexture(GL gl, String resourceName, int target,
			int dstPixelFormat, int minFilter, int magFilter, boolean wrap)
			throws IOException {

		BufferedImage image = loadImage(resourceName);

		int textureID = createTextureID(gl);

		gl.glBindTexture(target, textureID);

		image = convertImageDisplayFormat(image);

		System.out.println("Loading texture " + resourceName + " to ID: "
				+ textureID);

		Texture texture = new Texture(image, target, textureID);

		uploadTexture(gl, texture, target, dstPixelFormat, minFilter,
				magFilter, wrap);
		return texture;
	}

	private BufferedImage convertImageDisplayFormat(BufferedImage image) {
		ColorModel colorModel = glColorModel;
		int components = 3;

		if (image.getColorModel().hasAlpha()) {
			colorModel = glAlphaColorModel;
			components = 4;
		}

		WritableRaster raster = Raster.createInterleavedRaster(
				DataBuffer.TYPE_BYTE, image.getWidth(), image.getHeight(),
				components, null);

		BufferedImage texImage = new BufferedImage(colorModel, raster, false,
				new Hashtable());

		texImage.getGraphics().drawImage(image, 0, 0, null);

		return texImage;
	}

	private BufferedImage loadImage(String ref) throws IOException {
		BufferedImage image = ImageIO.read(new File(ref));
		AffineTransform tx = AffineTransform.getScaleInstance(1, -1);
		tx.translate(0, -image.getHeight());
		AffineTransformOp op = new AffineTransformOp(tx,
				AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
		image = op.filter(image, null);

		return image;
	}
}
